#include "MoveJointsToPosition.h"

#include <RobotAPI/libraries/SimpleTrajectory/Trajectory.h>

namespace armarx::control::skills::skills
{

    MoveJointsToPosition::MoveJointsToPosition(const Remote& r) :
        ::armarx::skills::SimpleSpecializedSkill<ParamType>(GetSkillDescription()), remote(r)
    {
    }

    std::map<std::string, float>
    MoveJointsToPosition::filterJointTargetValues(
        const NameValueMap& jointsTargetValues,
        const std::vector<std::string>& disabledJoints) const
    {
        std::map<std::string, float> filtered;
        for (const auto& [jointName, v] : jointsTargetValues)
        {
            if (std::find(disabledJoints.begin(), disabledJoints.end(), jointName) !=
                disabledJoints.end())
            {
                ARMARX_INFO << "Disable joint " << jointName
                            << " because it was explictly disabled in skill params.";
                continue;
            }
            filtered[jointName] = v;
        }
        return filtered;
    }

    ::armarx::skills::Skill::ExitResult
    MoveJointsToPosition::exit(const SpecializedExitInput& in)
    {
        auto filtered = this->filterJointTargetValues(in.parameters.jointsTargetValues,
                                                      in.parameters.disabledJoints);

        auto jointNames = simox::alg::get_keys(filtered);

        remote.kinematicUnit->releaseJoints(jointNames);
        return {::armarx::skills::TerminatedSkillStatus::Succeeded};
    }

    ::armarx::skills::Skill::MainResult
    MoveJointsToPosition::main(const SpecializedMainInput& in)
    {
        auto filtered = this->filterJointTargetValues(in.parameters.jointsTargetValues,
                                                      in.parameters.disabledJoints);

        auto jointNames = simox::alg::get_keys(filtered);

        // get VirtualRobot
        const VirtualRobot::RobotPtr virtualRobot =
            remote.virtualRobotReader.getRobot(remote.robotName);
        if (not virtualRobot)
        {
            ARMARX_ERROR << "Could not get robot '" << remote.robotName << "'";
            return MakeFailedResult();
        }

        // check jointsTargetValues are valid (joints exist, target values are within limits)
        if (not checkJointsTargetValuesValid(in.parameters.jointsTargetValues, virtualRobot))
        {
            return MakeFailedResult();
        }

        // request joints
        ARMARX_INFO << "Requesting joints";
        try
        {
            remote.kinematicUnit->requestJoints(jointNames);
        }
        catch (const armarx::KinematicUnitUnavailable& e)
        {
            ARMARX_IMPORTANT << "Failed to request joints. Terminating. Joints unavailable:"
                             << e.nodeOwners;
            return MakeFailedResult();
        }

        // move joints
        ARMARX_INFO << "Moving joints to " << filtered;

        switchControlMode(filtered, ePositionControl);

        if (in.parameters.inSimulation)
        {
            // trajectories from joint value updates through the KinematicUnit are not simulated
            // -> simulate manually
            moveJointsSimulation(filtered,
                                 in.parameters.simulationJointSpeed,
                                 in.parameters.simulationJointSpeedOverride,
                                 virtualRobot);
        }
        else
        {
            moveJoints(
                filtered, in.parameters.accuracy, in.parameters.accuracyOverride, virtualRobot);
        }
        return MakeSucceededResult();
    }

    bool
    MoveJointsToPosition::checkJointsTargetValuesValid(const NameValueMap& jointsTargetValues,
                                                       const VirtualRobot::RobotPtr& virtualRobot)
    {
        for (const auto& [jointName, targetValue] : jointsTargetValues)
        {
            if (not virtualRobot->hasRobotNode(jointName))
            {
                ARMARX_ERROR << "Robot has no joint '" << jointName << "'";
                return false;
            }
            const VirtualRobot::RobotNodePtr robotNode = virtualRobot->getRobotNode(jointName);
            if (not robotNode->checkJointLimits(targetValue))
            {
                ARMARX_ERROR << "Target value '" << targetValue << "' for joint '" << jointName
                             << "' out of limits (" << robotNode->getJointLimitLo() << " - "
                             << robotNode->getJointLimitHi() << ")";
                return false;
            }
        }
        return true;
    }

    void
    MoveJointsToPosition::switchControlMode(const NameValueMap& jointsTargetValues,
                                            const ControlMode controlMode)
    {
        armarx::NameControlModeMap modeMap;
        for (const auto& [jointName, _] : jointsTargetValues)
        {
            modeMap[jointName] = controlMode;
        }
        remote.kinematicUnit->switchControlMode(modeMap);
    }

    void
    MoveJointsToPosition::moveJoints(const NameValueMap& jointsTargetValues,
                                     const float accuracyDefault,
                                     const NameValueMap& accuracyOverride,
                                     const VirtualRobot::RobotPtr& virtualRobot)
    {
        remote.kinematicUnit->setJointAngles(jointsTargetValues);

        // wait until target values are reached
        ARMARX_INFO << "Waiting until target values are reached";

        const Clock clock;
        const Duration sleepTime = Duration::MilliSeconds(50);

        auto do_if_terminate_and_not_reached = [&]()
        {
            ARMARX_INFO << "Skill got aborted, stopping movement";
            ARMARX_CHECK(
                remote.virtualRobotReader.synchronizeRobot(*virtualRobot, armarx::DateTime::Now()))
                << "Robot state synchronization failed";

            // stop movement
            switchControlMode(jointsTargetValues, eVelocityControl);

            NameValueMap targetJointVelocities;
            for (auto& [jointName, _] : jointsTargetValues)
            {
                targetJointVelocities[jointName] = 0;
            }
            remote.kinematicUnit->setJointVelocities(targetJointVelocities);

            throwIfSkillShouldTerminate(); // throws exception

            ARMARX_INFO << "Target values reached";
        };

        bool reached = false;
        while (not reached)
        {
            throwIfSkillShouldTerminate(do_if_terminate_and_not_reached);
            ARMARX_CHECK(
                remote.virtualRobotReader.synchronizeRobot(*virtualRobot, armarx::DateTime::Now()))
                << "Robot state synchronization failed";

            reached = true;
            for (const auto& [jointName, targetValue] : jointsTargetValues)
            {
                const float currentValue = virtualRobot->getRobotNode(jointName)->getJointValue();
                const float accuracy = accuracyOverride.count(jointName) > 0
                                           ? accuracyOverride.at(jointName)
                                           : accuracyDefault;

                if (std::abs(targetValue - currentValue) > accuracy)
                {
                    reached = false;
                    ARMARX_VERBOSE << deactivateSpam(1) << "Delta for joint '" << jointName
                                   << "' is: " << std::abs(targetValue - currentValue) << " > "
                                   << accuracy << "(accuracy)";
                    break;
                }
            }
            clock.waitFor(sleepTime);
        }
    }

    void
    MoveJointsToPosition::moveJointsSimulation(const NameValueMap& jointsTargetValues,
                                               const float speedDefault,
                                               const NameValueMap& speedOverride,
                                               const VirtualRobot::RobotPtr& virtualRobot)
    {
        ARMARX_CHECK(
            remote.virtualRobotReader.synchronizeRobot(*virtualRobot, armarx::DateTime::Now()))
            << "Robot state synchronization failed";

        //build trajectory
        armarx::trajectory::Trajectory trajectory;
        double maxDuration = 0;

        for (const auto& [jointName, targetValue] : jointsTargetValues)
        {
            const float speed =
                speedOverride.count(jointName) > 0 ? speedOverride.at(jointName) : speedDefault;

            const float currentValue = virtualRobot->getRobotNode(jointName)->getJointValue();
            const double duration = std::abs(targetValue - currentValue) / speed;

            armarx::trajectory::VariantTrack& track = trajectory.addTrack(jointName);
            track[0] = currentValue;
            track[duration] = targetValue;

            maxDuration = std::max(maxDuration, duration);
        }

        ARMARX_INFO << "Simulating trajectory (duration: " << maxDuration << "s)";

        // run trajectory
        const Clock clock;
        const Duration sleepTime = Duration::MilliSeconds(10);

        const armarx::DateTime start = armarx::DateTime::Now();
        double t = 0;
        while (t < maxDuration)
        {
            throwIfSkillShouldTerminate();
            t = std::min((armarx::DateTime::Now() - start).toSecondsDouble(), maxDuration);

            NameValueMap nextValues;
            for (const auto& [name, _] : jointsTargetValues)
            {
                const float value = std::get<float>(trajectory[name].at(t));
                nextValues[name] = value;
            }

            remote.kinematicUnit->setJointAngles(nextValues);

            clock.waitFor(sleepTime);
        }
        ARMARX_INFO << "Target values reached";
    }

} // namespace armarx::control::skills::skills
